/* -*- Mode: C; c-file-style: "bsd" -*- */

/*  "hash cash" mint and remailer plug in for purposes of
 *  remailer postage (or any other use you can think of)
 * 
 *  hash cash is payment in burnt CPU cycles by calculating n-bit
 *  partial hash collisions
 *
 *  see:         http://www.cypherspace.org/hashcash/
 *
 */

#if defined( THINK_C )
    #include <console.h>
    #include <unix.h>
#else
    #include <unistd.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "db.h"
#include "timer.h"

#if defined( OPENSSL )
    #include <openssl/sha.h>
    #define SHA1_ctx SHA_CTX
    #define SHA1_Final( x, digest ) SHA1_Final( digest, x )
#else
    #include "sha1.h"
#endif

void random_start( word32*, word32*, word32*, word32* );
void usage( void );
word32 find_collision( char day[ 5 ], char*, int, char*, word32, char* );
double estimate_time( int );
double expected_tries( int );
double usecs_for_n_hashes( word32 );
int still_valid( char*, int );

#define GROUP_SIZE ULONG_MAX
#define GROUP_DIGITS 8
#define GFORMAT "%08x"

#if 0
/* smaller base for debugging */
#define GROUP_SIZE 255
#define GROUP_DIGITS 2
#define GFORMAT "%02x"
#endif

/* a 120Mhz 486 does 28000 per second

   do a special case for tick = 1 second, say 5 seconds
   so 131071 tries
*/

#define DEFAULT_TRY ( ( tick_us() == 1000000 ) ? 131071 : 16383 )

const char* database_filename = "hashcash.db";

word32 hashes_per_sec( void );

int validity_period = 28; 
int bits = 20;
char today[ 6 ];
word32 today_num;

#ifndef SPARSE
int main( int argc, char* argv[] )
{
    int tries;
    int i;
    word32 i0, i1, i2;
    word32 ran0, ran1, ran2, ran3;
    int done0 = 0, done1 = 0, done2 = 0;
    int collision_bits;
    int arguments = 0;
    int time_flag = 0;
    int bits_flag = 0;
    int validity_flag = 0;
    int database_flag = 0;
    int name_flag = 0;
    int check_flag = 0;
    int valid;
    int in_db;
    int gig;
    int expired;
    int seconds;
    int found;
    char collision[ 100 ];
    char name[ 100 ];
    char check[ 100 ];
    char counter[ 50 ];
    char* counter_ptr;
    double tries_expected;
    double tries_taken;
    double time_estimated;
    double taken;
    double count;

#ifdef THINK_C
    argc=ccommand(&argv);
#endif

    for ( i = 1; i < argc; i++ )
    {
	if ( argv[ i ][ 0 ] == '-' )
	{
	    switch ( argv[ i ][ 1 ] )
	    {
	    case 't':
		time_flag = 1;
		break;
	    case 'd':
		database_flag = 1;
		break;
	    case 'f':
		database_filename = argv[ i ] + 2;
		break;
	    case '0': case '1': case '2': case '3': case '4': 
	    case '5': case '6': case '7': case '8': case '9':
		bits = atoi( argv[ i ] + 1 );
		bits_flag = 1;
		if ( bits < 0 ) { usage(); }
		break;
	    default:
		usage();
	    }
	}
	else
	{
	    switch ( arguments )
	    {
	    case 0:
		strncpy( name, argv[ i ], 100 );
		name_flag = 1;
		break;
	    case 1:
		strncpy( check, argv[ i ], 100 );
		check_flag = 1;
		break;
	    case 2:
		validity_period = atoi( argv[ i ] );
		validity_flag = 1;
		if ( validity_period < 0 ) { usage(); }
		break;
	    default:
		usage();
		break;
	    }
	    arguments++;
	}
    }

    if ( ! time_flag && ! name_flag )
    {
	usage();
    }

    today_num = ( time( 0 ) / 3600 ) / 24;
    sprintf( today, "%05d", today_num );

    if ( check_flag )
    {
	/* database_flag only meaningful if check_flag set */

	valid = still_valid( check, validity_period );

	if ( validity_flag )
	{
	    printf( "validity: " );
	    if ( valid == 0 )
	    {
		printf( "expired\n" );
	    }
	    else if ( valid > 0 )
	    {
		printf( "%d days remaining\n", valid );
	    }
	    else
	    {
		printf( "forever\n" );
	    }
	}

	collision_bits = count_collision_bits( name, check );

	printf( "X-Hashcash: %d bits\n", collision_bits );
	if ( bits_flag )
	{
	    printf( "required: %d bits\n", bits );
	    if ( collision_bits < bits )
	    {
		printf( "rejected: too few bits\n" );
	    }
	}

	if ( database_flag )	/* if we're checking the database */
	{
	    db_open( database_filename );
	    in_db = db_in( check ); /* if it's in there, it's a failure */
	    printf( "database: " );
	    if ( in_db )
	    {
		printf( "double spent\n" );
		db_close();
		exit( 1 );
	    }
	    else
	    {
		printf( "not double spent\n" );
		if ( validity_flag ) /* validity check */
		{
		    if ( valid )
		    {
			if ( bits_flag ) /* check num of bits */
			{
			    if ( collision_bits >= bits )
			    {
				db_add( check, validity_period );
				db_close();
				exit( 0 );
			    }
			}
			else		/* no bit check */
			{
			    db_add( check, validity_period );
			    db_close();
			    exit( 0 );
			}
		    }
		    db_close();
		    exit( 1 );	/* else fail */
		}
		else		/* no validity check */
		{
		    if ( bits_flag ) /* check num of bits */
		    {
			if ( collision_bits >= bits )
			{
			    db_add( check, validity_period );
			    db_close();
			    exit( 0 );
			}
		    }
		    else		/* no bit check */
		    {
			db_add( check, validity_period );
			db_close();
			exit( 0 );
		    }
		    db_close();
		    exit ( 1 );
		}
	    }
	    db_close();
	}
	if ( !bits_flag )
	{
	    if ( !validity_flag || (validity_flag && valid) ) 
	    {
		exit ( 0 );		/* if target not given */
	    }
	}
	else 
	{
	    if ( collision_bits >= bits )
	    {
		if ( !validity_flag || (validity_flag && valid) ) 
		{
		    exit( 0 );		/* if target given and met */
		}
	    }
	}
	exit( 1 );		/* target given and failed */
    }

    printf( "speed: %lu hashes per sec\n", hashes_per_sec() );
    if ( bits_flag )
    {
	printf( "find: %d bit partial sha1 collision\n", bits );
	time_estimated = estimate_time( bits );
	printf( "estimate: %.0lf seconds\n", time_estimated );
	tries_expected = expected_tries( bits );
	printf( "expected: %.0lf tries = 2 ^ %d tries\n", tries_expected, bits );
    }
    fflush( stdout );
    if ( time_flag )
    {
	exit( 0 );		/* don't actually calculate it */
    }

    if ( ! name_flag )
    {
	usage();
    }

    memset( counter, '\0', 37 );

    random_start( &ran0, &ran1, &ran2, &ran3 );

    timer_start();
   
    counter_ptr = counter;
    found = 0;

    for ( i0 = 0; !found; )
    {

	sprintf( counter_ptr, GFORMAT GFORMAT, ( i0 + ran0 ) & GROUP_SIZE,
		 ran1 & GROUP_SIZE );
	found = find_collision( today, name, bits, collision,
				GROUP_SIZE, counter_ptr );
	if ( !found )
	{
	    i0 = ( i0 + 1 ) & GROUP_SIZE;
	    if ( i0 == 0 )	/* ie if wrapped */
	    { 
		fprintf( stderr, "failed to find collision in 2^64 tries!\n" );
		exit( 1 );
		/* shouldn't get here without trying  */
		/* for a very long time */
	    }
	}
    }

    count = 0;
    printf( "tries: " );
    counter_ptr = counter;
   
    tries_taken = ( (double) i0 ) * ULONG_MAX + ( (double) found );

    printf( "%.0lf", tries_taken );
    printf( " %.2lf x expected\n", tries_taken / tries_expected );

    timer_stop();
    taken = timer_time() / 1000000.0;
    printf( "time: %.0lf seconds\n", taken );
    putchar( '\n' );
    printf( "X-Hashcash: %s\n", collision );

    putchar( '\n' );

    return 0;
}

word32 hashes_per_s = 0;

word32 hashes_per_sec( void )
{
    double usecs;
    if ( hashes_per_s == 0 )
    {
	usecs = usecs_for_n_hashes( DEFAULT_TRY );
	hashes_per_s = (word32) ( ( ( (double)DEFAULT_TRY ) / usecs ) 
				  * 1000000.0 );
    }
    return hashes_per_s;
}

double expected_tries( int b )
{
    double expected_hashes = 1.0;
   
    while ( b > 31 )
    {
	expected_hashes *= 1UL << 31;
	b -= 31;
    }

    expected_hashes *= 1UL << b;
    return expected_hashes;
}

double estimate_time( int b )
{
    return expected_tries( b ) / hashes_per_sec();
}

double usecs_for_n_hashes( word32 try )
{
    timer start;
    timer end;
    byte collision[ 20 ];
    char counter[ 50 ];
    double elapsed;

    timer_read( &start );

    sprintf( counter, GFORMAT GFORMAT, 0, 0 );
    find_collision( "09856", "flame", 25, collision, try, counter );

    timer_read( &end );

    elapsed = timer_interval( &start, &end );

    return elapsed;
}
#endif SPARSE

int still_valid( char* check, int validity_period )
{
    char day[ 6 ];
    word32 day_num;
    word32 today_num = ( time( 0 ) / 3600 ) / 24;

    if ( validity_period == 0 )	/* for ever */
    {
	return -1;   /* infinity */
    }

    strncpy( day, check, 5 );
    day[ 5 ] = '\0';
    day_num = atoi( day );

    if ( day_num > today_num ) return 0;
    if ( day_num + validity_period > today_num )
    {				/* valid return days left */
	return day_num + validity_period - today_num;
    }
    else return 0;
}

int count_collision_bits( char* name, char* check )
{
    char target[ 100 ];
    char try[ 100 ];
    SHA1_ctx ctx;
    byte target_digest[ 20 ];
    byte check_digest[ 20 ];
    int i;
    int last;
    int collision_bits;

    strncpy( target, check, 5 );
    target[ 5 ] = ':';
    target[ 6 ] = '\0';
    strncat( target, name, 95 );

    SHA1_Init( &ctx );
    SHA1_Update( &ctx, target, strlen( target ) );
    SHA1_Final( &ctx, target_digest );
      
    SHA1_Init( &ctx );
    SHA1_Update( &ctx, check, strlen( check ) );
    SHA1_Final( &ctx, check_digest );
   
    for ( i = 0; i < 20 && check_digest[ i ] == target_digest[ i ]; i++ )
    {
    }

    last = i;
    collision_bits = 8 * i;

#define bit( n, c ) (((c) >> (7 - (n))) & 1)

    for ( i = 0; i < 8; i++ )
    {
	if ( bit( i, check_digest[ last ] ) == bit( i, target_digest[ last ] ) )
	{
	    collision_bits++;
	}
	else
	{
	    break;
	}
    }
    return collision_bits;
}

void random_start( word32* w1, word32* w2, word32* w3, word32* w4)
{
    /* kludge, might want to improve this, if collision collisions happen */

    timer t;
    word32 start;
    byte digest[ 20 ];
    word32* wdigest = (word32*)digest;

    SHA1_ctx ctx;

    SHA1_Init( &ctx );

    timer_read( &t );

    SHA1_Update( &ctx, (void*)&t, sizeof( timer ) );
    SHA1_Final( &ctx, digest );

    *w1 = wdigest[0];
    *w2 = wdigest[1];
    *w3 = wdigest[2];
    *w4 = wdigest[3];

    //   *w1 = ctx.H[ 0 ];
    //   *w2 = ctx.H[ 1 ];
    //   *w3 = ctx.H[ 2 ];
    //   *w4 = ctx.H[ 3 ];
}

word32 find_collision( char day[ 5 ], char* name, int bits, 
		       char* collision, word32 tries, char* counter )
{
    char* hex = "0123456789abcdef";
    char target[ 100 ];
    char try[ 100 ];
    char* changing_part_of_try;
    SHA1_ctx ctx;
    SHA1_ctx precomputed_ctx;
    word32 i;
    int j;
    word32 trial;
    word32 tries2;
    int len;
    int counter_len;
    int try_len;
    int try_strlen;
    byte target_digest[ 20 ];
    byte try_digest[ 20 ];
    int partial_byte = bits & 7;
    int check_bytes;
    int partial_byte_index;
    int partial_byte_mask;
    char last_char;
    int last_digit;
   
    memcpy( target, day, 5 );
    target[ 5 ] = ':';
    target[ 6 ] = '\0';
    strncat( target, name, 94 );

    counter_len = strlen( counter ) - GROUP_DIGITS;
    sscanf( counter + counter_len, "%08x", &trial );
    trial -= trial % 16;		/* lop off last digit */

    if ( partial_byte )
    {
	partial_byte_index = bits / 8;
	partial_byte_mask = ~ (( 1 << (8 - (bits & 7))) -1 );
	check_bytes = partial_byte_index + 1;
    }
    else
    {
	check_bytes = bits / 8;
    }

    SHA1_Init( &ctx );
    SHA1_Update( &ctx, target, strlen( target ) );
    SHA1_Final( &ctx, target_digest );
    //   memcpy( target_digest, SHA1_digest( &ctx ), SHA1_DIGEST_BYTES );

    if ( partial_byte )
    {
	target_digest[ partial_byte_index ] &= partial_byte_mask;
    }

    strncpy( try, target, 100 );
    strncat( try, ":", 1 );
    strncat( try, counter, counter_len );
    try_len = strlen( try );

/* length of try is fixed, GFORMAT is %08x, so move strlen outside loop */

    changing_part_of_try = try + try_len;
    sprintf( changing_part_of_try, GFORMAT, 0 );
    try_strlen = strlen( try );

/* part of the ctx context can be precomputed as not all of the
   message is changing
*/

    tries2 = (int) ( (double) tries / 16.0 + 0.5 );
    for ( i = 0; i < tries2; i++, trial = (trial + 16) & GROUP_SIZE )
    {
/* move precompute closer to the inner loop to precompute more */

	SHA1_Init( &precomputed_ctx );

	sprintf( changing_part_of_try, GFORMAT, trial );

	SHA1_Update( &precomputed_ctx, try, try_strlen - 1 );

	for ( j = 0; j < 16; j++ )
	{
#if defined( DEBUG )
	    printf( "try: %s\n", try );
#endif
	    memcpy( &ctx, &precomputed_ctx, sizeof( SHA1_ctx ) );
	    last_char = hex[ j ];
	    SHA1_Update( &ctx, &last_char, 1 );
	    SHA1_Final( &ctx, try_digest );
	    //	 memcpy( try_digest, SHA1_digest( &ctx ), SHA1_DIGEST_BYTES );
	    if ( bits > 7 )
	    {
		if ( try_digest[ 0 ] != target_digest[ 0 ] )
		{
		    continue;
		}
	    }
	    if ( partial_byte )
	    {
		try_digest[ partial_byte_index ] &= partial_byte_mask;
	    }
	    if ( memcmp( target_digest, try_digest, check_bytes ) == 0 )
	    {
		changing_part_of_try[ GROUP_DIGITS - 1 ] = hex[ j ];
		strcpy( collision, try );
		return i * 16 + j + 1;
	    }
	}
    }
    return 0;
}

#ifndef SPARSE
void usage( void )
{
    fprintf( stderr, "usage: hashcash [-t] [-d] [-fdbfile] [-bits] string [check [days]]\n\n" );
    fprintf( stderr, "\t-t\ttime only (don't find collision)\n" );
    fprintf( stderr, "\t-d\tuse database (for double spending detection)\n" );
    fprintf( stderr, "\t-f\tuse filename dbfile for database\n" );
    fprintf( stderr, "\t-bits\tfind partial hash collision of length bits\n" );
    fprintf( stderr, "\tstring\tis the string to find a partial hash collision on\n" );
    fprintf( stderr, "\tcheck\treport on collision check (with respect to string)\n" );
    fprintf( stderr, "\tdays\treport on cash with days validity period from minting\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "examples:\n" );
    fprintf( stderr, "\thashcash -20 flame                    # mint 20 bit collision\n" );
    fprintf( stderr, "\thashcash flame 09948flame3543         # examine collision\n" );
    fprintf( stderr, "# examine, also keep double spend database, validity 28 days\n" );
    fprintf( stderr, "\thashcash -d flame 09948flame3543 28\n" );
    fprintf( stderr, "\n" );
   
    fprintf( stderr, "see http://www.cypherspace.org/hashcash/ for more detail\n");
    exit( 1 );
}
#endif
